#!/usr/bin/env perl

use strict;
use warnings;
use Fcntl;
use File::Sync;
use IO::File;

my $cmdname = $0; $cmdname =~ s{.*/}{};
my $usage = "Usage: $cmdname .../cl.h opencl-errors.cpp";
die $usage unless @ARGV == 2;
my $header_file = $ARGV[0];
my $output_file = $ARGV[1];

my $in = new IO::File $header_file
   or die "$cmdname: unable to open $header_file: $!\n";
my $in_error_codes = 0;
my @codes; my $last_value;
while (<$in>) {
   next unless my ($code, $value) =
      m{^\#\s*define\s+(CL_\w+)\s+(-?\d+)\s*$};
   unless (defined $value &&
	 $value <= 0 && (!defined($last_value) || $value < $last_value)) {
      $in_error_codes = 0; next;
   }
   $in_error_codes = 1 if ($code eq "CL_SUCCESS" && $value == 0);
   next unless $in_error_codes;
   push(@codes, {symbol => $code, value => $value});
   $last_value = $value;
}
$in->close;

my $tmpfile = $output_file . ".TMP";
my $out = new IO::File $tmpfile, O_WRONLY|O_CREAT|O_EXCL
   or die "$cmdname: unable to create $tmpfile: $!\n";

print $out <<"END_DATA";
#include "opencl.hpp"
static struct {
   cl_int error;
   const char* symbol;
   cl_int block_len;
} error_codes[] = {
END_DATA
foreach my $index (0..$#codes) {
   my $code = $codes[$index];
   my $value = $code->{value};
   my $block_len = 0;
   my $index2 = $index + 1;
   while ($index2 <= $#codes &&
	 $value - $codes[$index2]->{value} == $index2 - $index) {
      ++$block_len; ++$index2;
   }
   print $out "   {", $value, ", \"",
      $code->{symbol}, "\", ", $block_len, "},\n";
}
print $out <<"END_TEXT";
};

const char* opencl_error_to_string(cl_int error) {
   if (error > 0 || error < $last_value) return 0;
   unsigned int index = 0;
   while (index < sizeof(error_codes)/sizeof(error_codes[0])) {
      cl_int value = error_codes[index].error;
      if (error == value) {
	 return error_codes[index].symbol;
      }
      if (error > value) return 0;
      cl_int block_len = error_codes[index].block_len;
      if (value - error <= block_len) {
	 index += value - error;
      } else {
	 index += block_len + 1;
      }
   }
   return 0;
}
END_TEXT
$out->flush or die "$cmdname: write error on $tmpfile: $!\n";
$out->fsync or die "$cmdname: write error on $tmpfile: $!\n";
$out->close or die "$cmdname: write error on $tmpfile: $!\n";
rename($tmpfile, $output_file) or
   die "$cmdname: unable to rename $tmpfile to $output_file: $!\n";
